#include <bits/stdc++.h>
#include <grpcpp/grpcpp.h>
#include<fstream>
#include "keyvaluestore.grpc.pb.h"
#define SERVERS "serverlist.txt"

using namespace std;

using grpc::Server;
using grpc::ServerAsyncResponseWriter;
using grpc::ServerBuilder;
using grpc::ServerCompletionQueue;
using grpc::ServerContext;
using grpc::Status;
using keyvaluestore::KeyValueServices;
using keyvaluestore::Info;
using keyvaluestore::Null;

ServerBuilder builder;
KeyValueServices::AsyncService service;
std::unique_ptr<Server> server;

enum RequestType {
    GETADDRESS,
    ADDADDRESS,
    UPDATEFINGERTABLES
};

class DNSData {
public:
    DNSData(KeyValueServices::AsyncService *service, ServerCompletionQueue *cq, RequestType reqType) : service(service), cq(cq), getAddressResponder(&context), addAddressResponder(&context), status(CREATE), reqType(reqType) {
        Proceed();
    }

    void Proceed() {
        if (status == CREATE) {
            status = PROCESS;
            if(reqType==GETADDRESS)
                service->RequestGETADDRESS(&context, &null, &getAddressResponder, cq, cq, this);
            else if(reqType==ADDADDRESS)
                service->RequestADDADDRESS(&context, &info, &addAddressResponder, cq, cq, this);
        }
        else if (status == PROCESS) {
            new DNSData(service, cq, reqType);
            if(reqType==GETADDRESS) {
                ifstream fin;
                int size=0;
                map<int,string> servers;
                fin.open(SERVERS);
                do {
                    string temp;
                    getline(fin,temp);
                    if(temp.size()==0)
                        break;
                    servers[size++]=temp;
                }while(fin);
                fin.close();
                if(size==0)
                    info.set_address("null");
                else {
                    int x=rand()%size;
                    info.set_address(servers.find(x)->second);
                }
                getAddressResponder.Finish(info,Status::OK,this);
            }
            else if(reqType==ADDADDRESS){
                ofstream fout;
                fout.open(SERVERS,ios::app);
                fout<<info.address()<<endl;
                fout.close();
                null.set_nothing(0);
                addAddressResponder.Finish(null,Status::OK,this);
            }
            else {
            }
            status = FINISH;
        }
        else {
            if(reqType==ADDADDRESS)
                print_servers_list();
            GPR_ASSERT(status == FINISH);
            delete this;
        }
    }

    void print_servers_list() {
        ifstream fin;
        map<int,string> servers;
        int size=0;
        fin.open(SERVERS);
        do {
            string temp;
            getline(fin,temp);
            if(temp.size()==0)
                break;
            servers[size++]=temp;
        }while(fin);
        fin.close();
        for(int i=0;i<size;i++)
            cout<<servers.find(i)->second<<endl;
    }

private:
    KeyValueServices::AsyncService *service;
    ServerCompletionQueue *cq;
    ServerContext context;
    Null null;
    Info info;
    ServerAsyncResponseWriter<Info> getAddressResponder;
    ServerAsyncResponseWriter<Null> addAddressResponder;
    enum CallStatus {
        CREATE,
        PROCESS,
        FINISH
    };
    CallStatus status;
    RequestType reqType;
};

int main(int argc,char **argv) {
    srand(time(0));
    string server_address("0.0.0.0:1234");
    builder.AddListeningPort(server_address, grpc::InsecureServerCredentials());
    builder.RegisterService(&service);
    unique_ptr<ServerCompletionQueue> comp_queue=builder.AddCompletionQueue();
    server = builder.BuildAndStart();
    cout<<"DNS SERVER COMES UP SUCCESSFULLY"<<endl;
    new DNSData(&service,comp_queue.get(),GETADDRESS);
    new DNSData(&service,comp_queue.get(),ADDADDRESS);
    new DNSData(&service,comp_queue.get(),UPDATEFINGERTABLES);
    void *tag;
    bool ok;
    while(true) {
        GPR_ASSERT(comp_queue->Next(&tag,&ok));
        GPR_ASSERT(ok);
        static_cast<DNSData *>(tag)->Proceed();
    }
}